#ifndef __SD_LIB_H__
#define __SD_LIB_H__

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <vector>
#include <string>

#include <stdio.h>
#include <glib.h>

#include "dictziplib.h"
#include "mapfile.h"
#include "lib_utils.h"


struct cacheItem{
  glong offset;
  gchar *data;
  inline cacheItem(void) : data(NULL){}
  inline ~cacheItem(){g_free(data);}
};

const int WORDDATA_CACHE_NUM = 10;
const int INVALID_INDEX=-100;

struct TDictInfo{
  std::string ifofilename;
  glong wordcount;
  std::string bookname;
  std::string author;
  std::string email;
  std::string website;
  std::string date;
  std::string description;
  gulong idxfilesize;
  std::string sametypesequence;
};

class DictBase{
public:
  DictBase();
  ~DictBase();
  gchar * GetWordData(glong idxitem_offset, glong idxitem_size);
  static bool get_dict_info(const gchar *ifofilename,  TDictInfo & dict_info, bool istreedict=false);
protected:
  gchar *sametypesequence;
  FILE *dictfile;
  dictData *dictdzfile;
private:
  struct cacheItem cache[WORDDATA_CACHE_NUM];
  gint cache_cur;	
};


struct TSeqBundle{
  gunichar val;
  gunichar dif;
  TSeqBundle(void){}
  TSeqBundle(gunichar v, gunichar d) : val(v), dif(d){}
};

typedef std::vector<TDictInfo> TDictInfoList;

class Dictionary : public DictBase{
private:
  glong wordcount;
  gchar *bookname;
  FILE *idxfile;
  union{gchar **wordlist;  glong *wordoffset;};
  bool need_free_wordoffset;
  union{gchar *idxdatabuffer;  glong cur_wordindex;};

  TMapFile map_wordoffset;
  GArray *alphabet_index;
  GArray *bundle_list;

  // The length of "word_str" should be less than 256. See src/tools/DICTFILE_FORMAT.
  gchar wordentry_buf[256]; 
  glong wordentry_offset;
  glong wordentry_size;
  const gchar *cache_dir;
  gchar *m_ifofilename;

  gboolean load_ifofile(const char *ifofilename, gulong *idxfilesize);
  void loadwordlist(void);	
  gboolean loadwordoffset(const char *idxfilename, gulong idxfilesize);
public:
  explicit Dictionary(const gchar *_cache_dir);
  ~Dictionary();
  gboolean load(const char *ifofilename);
  inline glong length() { return(wordcount); }
  inline gchar* GetBookname() { return(bookname); }
  gboolean Lookup(const char* sWord,glong *pIndex);
  gboolean LookupWithRule(GPatternSpec *pspec,glong *aIndex,int iBuffLen);
  gchar * GetWord(glong index);
  gchar * GetWordData(glong index);
  const gchar *get_ifofilename(void) const{return m_ifofilename;}
};

//============================================================================
// struct
struct Fuzzystruct {
  char * pMatchWord;
  int iMatchWordDistance;
};

typedef void TProgressFunc(void);

const int MAX_FUZZY_MATCH_ITEM=100;
const int MAX_FUZZY_DISTANCE= 3; // at most MAX_FUZZY_DISTANCE-1 differences allowed when find similar words
const int MAX_FLOAT_WINDOW_FUZZY_MATCH_ITEM=5;
const int MAX_MATCH_ITEM_PER_LIB=100;

class Library{
private:
  std::vector<Dictionary*> oLib; // word library.
 
  int iMaxFuzzyDistance;
  gchar *cache_dir;
  
  static int FuzzystructCompare(const void * s1, const void * s2);
  static int MatchWordCompare(const void * s1, const void * s2);

  void LoadDir(gchar *dirname, const GSList *order_list, const GSList *disable_list);
  gboolean SimpleLookupWord(const gchar* sWord,glong& iWordIndex,int iLib);
  inline gboolean LookupWord(const gchar* sWord,glong& iWordIndex,int iLib){return (oLib[iLib]->Lookup(sWord, &iWordIndex));}
  gboolean LookupSimilarWord(const gchar* sWord,glong& iWordIndex,int iLib);
  inline gboolean LookdupWordsWithRule(GPatternSpec *pspec,glong* aiIndexes, int iLen, int iLib){  return (oLib[iLib]->LookupWithRule(pspec,aiIndexes,iLen));}
public:
  explicit Library(gchar *_cache_dir);
  ~Library();
  void Load(const GSList *order_list, const GSList  *disable_list, const gchar *data_dir=NULL);
  void ReLoad(const GSList *order_list, const GSList *disable_list, const gchar *data_dir=NULL);
  inline glong iLength(int iLib) const{return (oLib[iLib])->length();} 
  inline const gchar* GetBookname(int iLib) const{return (oLib[iLib])->GetBookname();}
  inline gint total_libs(void) const{return oLib.size();}
  inline const gchar* get_ifofilename(gint iDict) const{return oLib[iDict]->get_ifofilename();}
  inline gchar *poGetWord(glong iIndex,int iLib){return( oLib[iLib]->GetWord(iIndex) );}
  inline gchar *poGetWordData(glong iIndex,int iLib){
    if(iIndex==INVALID_INDEX)
      return NULL;
    return oLib[iLib]->GetWordData(iIndex);
  }
 
  gchar * poGetCurrentWord(glong * iCurrent);
  gchar * poGetNextWord(const gchar *word,glong * iCurrent);
  gchar * poGetPreWord(glong * iCurrent);

  gboolean SimpleLookup(const gchar* sWord, gchar ** & ppWord, 
			gchar ** & ppWordData, gchar * & SearchWord);
  gboolean SimpleLookup(const gchar* sWord, glong* piIndex,
			gchar ** & ppWord, gchar ** & ppWordData,
			gboolean piIndexValid, 
			gboolean bTryMoreIfNotFound);

  gboolean LookupWithFuzzy(const gchar *sWord, Fuzzystruct * & oFuzzystruct, TProgressFunc progress_func=NULL);
  gboolean LookupWithFuzzy(const gchar *sWord, gchar *** & pppWord, gchar *** & pppWordData, gchar ** & ppOriginWord, gint & count, TProgressFunc progress_func=NULL);
  gint LookupWithRule(const gchar *word, gchar ** & ppMatchWord, TProgressFunc progress_func=NULL);
  static void load_dict_info_list_from_dir(const gchar *dir_name, TDictInfoList & dict_info_list);
  static void get_dict_info_list(TDictInfoList & dict_info_list, const gchar *data_dir=NULL);
};


#endif
